#---------------------------------------
# BNCmatmul.inc for including in Makefile
# Copyright (c) 2014-2016 Tomonori Kouya
#---------------------------------------

#----
# Version number
#----
# Version 0.1: first release: implementation of single-threaded strassen multiplication based on IEEE754 double and MPFR
#VERSION=0.1
# Version 0.2: Appended Parallelization and DD and QD prec. routines
VERSION=0.2

#----
# BNCmatmul director
#----
BNCMATMUL_PREFIX=bncmatmul-$(VERSION)

#----
# Name of Library
#----
BNCMATMUL_LIBNAME_SUFFIX = $(BNCMATMUL_PREFIX)
BNCMATMUL_LIBNAME_OMP_SUFFIX = $(BNCMATMUL_PREFIX)-omp
BNCMATMUL_LIBNAME_QD_SUFFIX = $(BNCMATMUL_PREFIX)-qd
BNCMATMUL_LIBNAME = lib$(BNCMATMUL_LIBNAME_SUFFIX).a
BNCMATMUL_LIBNAME_OMP = lib$(BNCMATMUL_LIBNAME_OMP_SUFFIX).a
BNCMATMUL_LIBNAME_QD = lib$(BNCMATMUL_LIBNAME_QD_SUFFIX).a
BNCMATMUL_LIBNAME_CC_SUFFIX = $(BNCMATMUL_PREFIX)_cc
BNCMATMUL_LIBNAME_CC_OMP_SUFFIX = $(BNCMATMUL_PREFIX)_cc-omp
BNCMATMUL_LIBNAME_CC = lib$(BNCMATMUL_LIBNAME_CC_SUFFIX).a
BNCMATMUL_LIBNAME_CC_OMP = lib$(BNCMATMUL_LIBNAME_CC_OMP_SUFFIX).a
BNCMATMUL_INC = -I./
BNCMATMUL_LIB = -L./ -l$(BNCMATMUL_LIBNAME_SUFFIX)
BNCMATMUL_LIB_OMP = -L./ -l$(BNCMATMUL_LIBNAME_OMP_SUFFIX) -l$(BNCMATMUL_LIBNAME_SUFFIX)
#BNCMATMUL_LIB = -L./ -l$(BNCMATMUL_LIBNAME_SUFFIX) -l$(BNCMATMUL_LIBNAME_SUFFIX)-qd
BNCMATMUL_LIB_CC = -L./ -l$(BNCMATMUL_LIBNAME_CC_SUFFIX)
BNCMATMUL_LIB_CC_OMP = -L./ -l$(BNCMATMUL_LIBNAME_CC_OMP_SUFFIX) -l$(BNCMATMUL_LIBNAME_CC_SUFFIX)

#----
# Bailey's qd lib
#----
QD_INC = -I/usr/local/include -DUSE_DD -DUSE_QD
QD_LIB = -L/usr/local/lib -lqd

#----
# GMP, MPFR and BNCpack
#----
GMP_INC = -I/usr/local/include
GMP_LIB_PATH = -L/usr/local/lib
GMP_LIB = -lgmp
MPFR_INC = $(GMP_INC)
MPFR_LIB = $(GMP_LIB_PATH) -lmpfr $(GMP_LIB)
BNC_INC = $(GMP_INC) -DUSE_GMP -DUSE_MPFR
BNC_LIB = $(GMP_LIB_PATH) -lbnc $(MPFR_LIB) -lm

#----
# Use GCC
#----
#CC = gcc
CC_ONLY = gcc
CC_OPT = 
CC = $(CC_ONLY) $(CC_OPT)
CC_DD = $(CC_ONLY) $(CC_OPT)
#CXX = g++
CXX_ONLY = g++
CXX_OPT = # -std=c++11
CXX = $(CXX_ONLY) $(CXX_OPT)
COMMON_INC = $(QD_INC) $(BNC_INC)
COMMON_LIB = $(BNC_LIB) -lpthread $(QD_LIB)
#----
# OpenMP options (Partially working)
#----
OPENMP = -fopenmp # gcc
OPENMPLIB = -lgomp # gcc + gomp

#----
# USE Intel C compiler + Math kernel
#----
#CC = icc
#CC = icc -g
#CC_ONLY = icc
#CC_OPT = -g -fp-model precise
#CC = $(CC_ONLY) $(CXX_OPT)
#CXX = icpc
#CXX_ONLY = icpc
#CXX_OPT = -fp-model precise
#CXX = $(CXX_ONLY) $(CXX_OPT)
#CXX = icpc -pch -pch-dir /home/tkouya/na/bncmatmul-0.2
#COMMON_INC = -I/opt/intel/mkl/include -DUSE_IMKL $(BNC_INC) $(QD_INC)
#COMMON_LIB = -L/opt/intel/mkl/lib/intel64 -lmkl_intel_lp64 -lmkl_sequential -lmkl_core -lpthread $(BNC_LIB) $(QD_LIB)
#----
# OpenMP options (Partially working)
#----
#OPENMP = -DUSE_OMP -openmp -openmp-report2 -O3 #intel
#OPENMP = -DUSE_OMP -openmp -openmp-report2 -O1 #intel
#OPENMPLIB = # intel

#----
# OpenMP common options (Partially working)
#----
# Number of threads
#OPENMP_NUM_THREADS = 1
#OPENMP_NUM_THREADS = 2
#OPENMP_NUM_THREADS = 3
#OPENMP_NUM_THREADS = 4
#OPENMP_NUM_THREADS = 5
#OPENMP_NUM_THREADS = 6
#OPENMP_NUM_THREADS = 7
OPENMP_NUM_THREADS = 8

#----
# AR
#---
AR = ar rc

#----
# NVCC of CUDA
#----
#NVCC = nvcc -g -arch=sm_30 -ccbin=$(CXX_ONLY) -Xcompiler='$(CXX_OPT)'
NVCC = nvcc -g -DCUDA_FMA -arch=sm_30 -ccbin=$(CXX_ONLY) -Xcompiler='$(CXX_OPT)'
GQD_INC_DIR = ./gqd_1_3/inc
GQD_INC = -I/usr/local/cuda/include -I$(GQD_INC_DIR)
#GQD_LIB = -L/usr/local/cuda/lib64 -lcudart_static -lcudart -lcudadevrt
GQD_LIB = -L/usr/local/cuda/lib64 -lcudart -lcudadevrt

#----
# Macros to compile programs with bncmatmul library
#----
INC = $(COMMON_INC) $(BNCMATMUL_INC)
LIB = $(BNCMATMUL_LIB) $(COMMON_LIB) 
LIB_OMP = $(BNCMATMUL_LIB_OMP) $(COMMON_LIB) 
LIB_CC = $(BNCMATMUL_LIB_CC) $(COMMON_LIB) 
LIB_CC_OMP = $(BNCMATMUL_LIB_CC_OMP) $(COMMON_LIB) 

#----
# Constants and Options for benchmark tests
#----
CACHE_SIZE = 8192 # KB
MPF_PREC = 1024 # in bits
OMP_NUM_THREADS = 4 # no.threads
#OMP_NUM_THREADS = 8 # no.threads
ROW_DIM = 128
MIN_DIM = 32 # minimal dimensions
MAX_SECONDS = 300 # Max. seconds to finalize benchmarking
LU_DIM = 1024

