/********************************************************************************/
/* bncomp.h: Parallelized Mutiple Precision Linear Computation Library          */
/*                                                             with OpenMP      */
/* Copyright (C) 2013-2015 Tomonori Kouya                                       */
/*                                                                              */
/* This program is free software: you can redistribute it and/or modify it      */
/* under the terms of the GNU Lesser General Public License as published by the */
/* Free Software Foundation, either version 3 of the License or any later       */
/* version.                                                                     */
/*                                                                              */
/* This program is distributed in the hope that it will be useful, but WITHOUT  */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License */
/* for more details.                                                            */
/*                                                                              */
/* You should have received a copy of the GNU Lesser General Public License     */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                              */
/********************************************************************************/
#include "bnc.h"
#include "ddlinear.h"

#ifndef __BNCOMP_H
#define __BNCOMP_H

#include <omp.h>

#ifdef __cplusplus
extern "C" {
#endif // __cplusplus

/* global variable _bncomp */
#define BNCOMP_MAX_NUM_THREADS 128
//#define BNCOMP_MAX_NUM_THREADS 1024

#ifdef _OPENMP
	static int _bncomp_num_threads = BNCOMP_MAX_NUM_THREADS;
	//static int _bncomp_num_tmp = MAX_NUM_BNCOMP_TMP;
	static int _bncomp_num_tmp = BNCOMP_MAX_NUM_THREADS;
#else // _OPENMP
	static int _bncomp_num_threads = 1;
	//static int _bncomp_num_tmp = MAX_NUM_BNCOMP_TMP;
	static int _bncomp_num_tmp = 1;
#endif // _OPENMP

// Global working variable for double precision arithmetic
//#ifndef _BNCOMP_DEFINE_BNCOMP_GVALS
//#define _BNCOMP_DEFINE_BNCOMP_GVALS

extern double  _bncomp_g_dval[BNCOMP_MAX_NUM_THREADS];
extern DVector _bncomp_g_dvec[BNCOMP_MAX_NUM_THREADS];
extern DMatrix _bncomp_g_dmat[BNCOMP_MAX_NUM_THREADS];

// Global working variables for multiple precision arithmetic
#ifdef USE_GMP
extern mpf_t     _bncomp_g_mpfval[BNCOMP_MAX_NUM_THREADS];
extern mpf_t     _bncomp_g_mpfval2[BNCOMP_MAX_NUM_THREADS];
extern MPFVector _bncomp_g_mpfvec[BNCOMP_MAX_NUM_THREADS];
extern MPFVector _bncomp_g_mpfvec2[BNCOMP_MAX_NUM_THREADS];
extern MPFVector _bncomp_g_mpfvec3[BNCOMP_MAX_NUM_THREADS];
extern MPFVector _bncomp_g_mpfvec4[BNCOMP_MAX_NUM_THREADS];
extern MPFMatrix _bncomp_g_mpfmat[BNCOMP_MAX_NUM_THREADS];
#endif // USE_GMP

/*
double  _bncomp_g_dval[BNCOMP_MAX_NUM_THREADS];
DVector _bncomp_g_dvec[BNCOMP_MAX_NUM_THREADS];
DMatrix _bncomp_g_dmat[BNCOMP_MAX_NUM_THREADS];

// Global working variables for multiple precision arithmetic
#ifdef USE_GMP
mpf_t     _bncomp_g_mpfval[BNCOMP_MAX_NUM_THREADS];
mpf_t     _bncomp_g_mpfval2[BNCOMP_MAX_NUM_THREADS];
MPFVector _bncomp_g_mpfvec[BNCOMP_MAX_NUM_THREADS];
MPFVector _bncomp_g_mpfvec2[BNCOMP_MAX_NUM_THREADS];
MPFVector _bncomp_g_mpfvec3[BNCOMP_MAX_NUM_THREADS];
MPFVector _bncomp_g_mpfvec4[BNCOMP_MAX_NUM_THREADS];
MPFMatrix _bncomp_g_mpfmat[BNCOMP_MAX_NUM_THREADS];
#endif // USE_GMP
*/

//#endif // _BNCOMP_DEFINE_BNCOMP_GVALS

// set_bncomp_num_threads(int num_threads)
int set_bncomp_num_threads(int num_threads);

// get number of threads in BNCOMP
int get_bncomp_num_threads(void);

/* initialize */
void _bncomp_init_g_d(int vec_dim, int mat_row_dim, int mat_col_dim);

/* finalize */
void _bncomp_free_g_d(void);

/* substitute vector blocks */
void _bncomp_subst_dvector_blocks(DVector ret[], DVector v[], long int num_blocks);

/* inner product of vector blocks */
double _bncomp_ip_dvector_blocks(DVector va[], DVector vb[], long int num_blocks);

/* norm2 of vector blocks */
double _bncomp_norm2_dvector_blocks(DVector v[], long int num_blocks);

/* ret := va + alpha * vb */
void _bncomp_add_cmul_dvector_blocks(DVector ret[], DVector va[], double alpha, DVector vb[], long int num_blocks);

/* ret := (A \otimes I) vb[] */
void _bncomp_kmul_dmatrixI_dvector_blocks(DVector ret[], DMatrix mat, DVector vb[], long int num_blocks);

/* ret := -vb */
void _bncomp_neg_dvector_blocks(DVector vb[], long int num_blocks);

/* norm_inf of vector blocks */
double _bncomp_normi_dvector_blocks(DVector v[], long int num_blocks);

/* ret := va - vb */
void _bncomp_sub_dvector_blocks(DVector ret[], DVector va[], DVector vb[], long int num_blocks);

/* ret := va + vb */
void _bncomp_add_dvector_blocks(DVector ret[], DVector va[], DVector vb[], long int num_blocks);

/* ret := sum^n_{i,j=1} aij^2 */
double _bncomp_sumsqr_dmatrix(DMatrix mat);

/* Frobenius Norm for tridiagonal block matrix */
double _bncomp_normf_dmatrix_tridiag_blocks(DMatrix aim1[], DMatrix aii[], DMatrix aip1[], long int num_blocks);

/* [ ret[0]              ]    [ b[0]              ]   [ aii [0] aip1[0]                                                   ]   [ x[0]              ] */
/* [ ret[1]              ]    [ b[1]              ]   [ aim1[0] aii [1] aip1[1]                                           ]   [ x[1]              ] */
/* [ ................... ] := [ ................. ] - [ ................................................................. ] * [ ................. ] */
/* [ ret[num_blocks - 2] ]    [ b[num_blocks - 2] ]   [    aim1[num_blocks - 3] aii [num_blocks - 2] aip1[num_blocsk - 2] ]   [ x[num_blocks - 2] ] */
/* [ ret[num_blocks - 1] ]    [ b[num_blocks - 1] ]   [                         aim1[num_blocks - 2] aii [num_blocks - 1] ]   [ x[num_blocks - 1] ] */
void _bncomp_residual_dmat_dvec_tridiag_blocks(DVector ret[], DVector b[], DMatrix aim1[], DMatrix aii[], DMatrix aip1[], DVector x[], long int num_blocks);

/* residual for band matrix */
void _bncomp_residual_dbmat_dvec_tridiag_blocks(DVector ret[], DVector b[], DBMatrix aim1[], DBMatrix aii[], DBMatrix aip1[], DVector x[], long int num_blocks);

/* substitute vector blocks */
void _bncomp_subst_fvector_dvec_blocks(FVector ret[], DVector v[], long int num_blocks);

/* substitute vector blocks */
void _bncomp_subst_dvector_fvec_blocks(DVector ret[], FVector v[], long int num_blocks);

/* substitution of tridiagonal block matrix */
void _bncomp_subst_fmatrix_dmat_tridiag_blocks(FMatrix aim1_f[], FMatrix aii_f[], FMatrix aip1_f[], DMatrix aim1[], DMatrix aii[], DMatrix aip1[], long int num_blocks);

/* ret[] := a[] * vb[] */
void _bncomp_mul_dtridiag_dvec_blocks(DVector ret[], DMatrix aim1[], DMatrix aii[], DMatrix aip1[], DVector vb[], DVector tmpv[], long int num_blocks);

/* ret[] := a_band[] * vb[] */
void _bncomp_mul_dbtridiag_dvec_blocks(DVector ret[], DBMatrix aim1[], DBMatrix aii[], DBMatrix aip1[], DVector vb[], DVector tmpv[], long int num_blocks);

/* norm_tol2_dvector */
double _bncomp_norm_tol2_dvector(DVector y_err, DVector y_new, DVector y_old, double rtol, double atol);

/* norm_tol2_dvector */
double _bncomp_norm2_diff_dvector(DVector y_new, DVector y_old, double rtol, double atol);

/* norm1 of vector blocks */
double _bncomp_norm1_dvector_blocks(DVector v[], long int num_blocks);

// normalize of vector
// kind of vector norm: 0...Infinity, 1...1norm, 2...euclid norm
// return value : ||v||
#define BNC_INFINITY_NORM	0
#define BNC_ONE_NORM		1
#define BNC_EUCLID_NORM		2
double _bncomp_norm_dvector(DVector vec, int kind_of_norm);

/* ret := val * vec */
void _bncomp_cmul_dvector_blocks(DVector ret[], double val, DVector vec[], long int num_blocks);

/* vec := val * vec */
void _bncomp_cmul2_dvector_blocks(DVector vec[], double val, long int num_blocks);

/* ||vec|| */
double _bncomp_norm_dvector_blocks(DVector vec[], long int num_blocks, int kind_of_norm);

/* ret := vec / ||vec|| */
double _bncomp_normalize_dvector_blocks(DVector ret[], DVector vec[], long int num_blocks, int kind_of_norm);

/* vec := vec / ||vec|| */
double _bncomp_normalize2_dvector_blocks(DVector vec[], long int num_blocks, int kind_of_norm);

/* BiCGSTAB for band matrix */
long int _bncomp_DBBiCGSTAB_triblock(DVector answer[], DBMatrix aim1[], DBMatrix aii[], DBMatrix aip1[], DVector b[], double reps, double aeps, long int maxtimes, long int num_blocks);

/* BiCGSTAB */
long int _bncomp_DBiCGSTAB_triblock(DVector answer[], DMatrix aim1[], DMatrix aii[], DMatrix aip1[], DVector b[], double reps, double aeps, long int maxtimes, long int num_blocks);

/* Left preconditioned BiCGSTAB for band matrix */
long int _bncomp_DBBiCGSTAB_triblock_irk(DVector answer[], DBMatrix aim1[], DBMatrix aii[], DBMatrix aip1[], DBMatrix hmat[], DVector b[], double reps, double aeps, long int maxtimes, long int num_blocks);

/* Left preconditioned BiCGSTAB */
long int _bncomp_DBiCGSTAB_triblock_irk(DVector answer[], DMatrix aim1[], DMatrix aii[], DMatrix aip1[], DMatrix hmat[], DVector b[], double reps, double aeps, long int maxtimes, long int num_blocks);

/* GMRES(m) for band matrix */
int _bncomp_DBGMRESm_triblock(DVector ret_x[], DBMatrix aim1[], DBMatrix aii[], DBMatrix aip1[], DVector vec_b[], int restart_times, int maxtimes, double rtol, double atol, long int num_blocks);

/* GMRES(m) */
int _bncomp_DGMRESm_triblock(DVector ret_x[], DMatrix aim1[], DMatrix aii[], DMatrix aip1[], DVector vec_b[], int restart_times, int maxtimes, double rtol, double atol, long int num_blocks);

/******/
/* DD */
/******/
/* c := a */
void _bncomp_subst_ddvector(DDVector c, DDVector a);

/* c = a + b */
void _bncomp_add_ddvector(DDVector c, DDVector a, DDVector b);

/* c = a - b */
void _bncomp_sub_ddvector(DDVector c, DDVector a, DDVector b);

/* c = val * a */
#ifdef __cplusplus
void _bncomp_cmul_ddvector(DDVector c, dd_real val, DDVector a);
#else // __cplusplus
void _bncomp_cmul_ddvector(DDVector c, double val[DDSIZE], DDVector a);
#endif // __cplusplus

/* (a, b) */
#ifdef __cplusplus
void _bncomp_ip_ddvector(dd_real *ret, DDVector a, DDVector b);
#else // __cplusplus
void _bncomp_ip_ddvector(double ret[DDSIZE], DDVector a, DDVector b);
#endif // __cplusplus

/* c := a + b */
void _bncomp_add_ddmatrix(DDMatrix c, DDMatrix a, DDMatrix b);

/* c := a - b */
void _bncomp_sub_ddmatrix(DDMatrix c, DDMatrix a, DDMatrix b);

/* c := sc * a */
#ifdef __cplusplus
void _bncomp_cmul_ddmatrix(DDMatrix c, dd_real sc, DDMatrix a);
#else // __cplusplus
void _bncomp_cmul_ddmatrix(DDMatrix c, double sc[DDSIZE], DDMatrix a);
#endif // __cplusplus

/* c = a * b */
void _bncomp_mul_ddmatrix(DDMatrix ret, DDMatrix a, DDMatrix b);

/* c := a */
void subst_ddmatrix(DDMatrix c, DDMatrix a);

/* c := I */
void _bncomp_setI_ddmatrix(DDMatrix c);

// set a zero matrix
//void set0_ddmatrix(DDMatrix mat)
void _bncomp_set0_ddmatrix(DDMatrix mat);

/* v := a * vb */
void _bncomp_mul_ddmatrix_ddvec(DDVector v, DDMatrix a, DDVector vb);

/* v := a^T * vb */
void _bncomp_mul_ddmatrixt_ddvec(DDVector v, DDMatrix a, DDVector vb);

/******/
/* QD */
/******/

/* c := a */
void _bncomp_subst_qdvector(QDVector c, QDVector a);

/* c = a + b */
void _bncomp_add_qdvector(QDVector c, QDVector a, QDVector b);

/* c = a - b */
void _bncomp_sub_qdvector(QDVector c, QDVector a, QDVector b);

/* c = val * a */
#ifdef __cplusplus
void _bncomp_cmul_qdvector(QDVector c, qd_real val, QDVector a);
#else // __cplusplus
void _bncomp_cmul_qdvector(QDVector c, double val[QDSIZE], QDVector a);
#endif // __cplusplus

/* (a, b) */
#ifdef __cplusplus
void _bncomp_ip_qdvector(qd_real *ret, QDVector a, QDVector b);
#else // __cplusplus
void _bncomp_ip_qdvector(double ret[QDSIZE], QDVector a, QDVector b);
#endif // __cplusplus

/* c := a + b */
void _bncomp_add_qdmatrix(QDMatrix c, QDMatrix a, QDMatrix b);

/* c := a - b */
void _bncomp_sub_qdmatrix(QDMatrix c, QDMatrix a, QDMatrix b);

/* c := sc * a */
#ifdef __cplusplus
void _bncomp_cmul_qdmatrix(QDMatrix c, dd_real sc, QDMatrix a);
#else // __cplusplus
void _bncomp_cmul_qdmatrix(QDMatrix c, double sc[QDSIZE], QDMatrix a);
#endif // __cplusplus

/* c = a * b */
void _bncomp_mul_qdmatrix(QDMatrix ret, QDMatrix a, QDMatrix b);

/* c := a */
void _bncomp_subst_qdmatrix(QDMatrix c, QDMatrix a);

/* c := I */
void _bncomp_setI_qdmatrix(QDMatrix c);

// set a zero matrix
//void set0_qdmatrix(QDMatrix mat)
void _bncomp_set0_qdmatrix(QDMatrix mat);

/* v := a * vb */
void _bncomp_mul_qdmatrix_qdvec(QDVector v, QDMatrix a, QDVector vb);

/* v := a^T * vb */
void _bncomp_mul_qdmatrixt_qdvec(QDVector v, QDMatrix a, QDVector vb);

/*******/
/* MPF */
/*******/
#ifdef USE_GMP

/* c := a */
void _bncomp_subst_mpfvector(MPFVector c, MPFVector a, long int dim);

/* c = a + b */
void _bncomp_add_mpfvector(MPFVector c, MPFVector a, MPFVector b, long int dim);

/* c = a - b */
void _bncomp_sub_mpfvector(MPFVector c, MPFVector a, MPFVector b, long int dim);

/* c = val * a */
void _bncomp_cmul_mpfvector(MPFVector c, mpf_t val, MPFVector a, long int dim);

// Matrix
/* c := a + b */
void _bncomp_add_mpfmatrix(MPFMatrix c, MPFMatrix a, MPFMatrix b);

/* c := a - b */
void _bncomp_sub_mpfmatrix(MPFMatrix c, MPFMatrix a, MPFMatrix b);

/* c := sc * a */
void _bncomp_cmul_mpfmatrix(MPFMatrix c, mpf_t sc, MPFMatrix a);

/* inner product of vector blocks */
void _bncomp_ip_mpfvector(mpf_t ret, MPFVector va, MPFVector vb);

/* c = a * b */
void _bncomp_mul_mpfmatrix(MPFMatrix c, MPFMatrix a, MPFMatrix b);

/* c := a */
void _bncomp_subst_mpfmatrix(MPFMatrix c, MPFMatrix a);

/* c := 0 */
void _bncomp_set0_mpfmatrix(MPFMatrix c);

/* c := I */
void _bncomp_setI_mpfmatrix(MPFMatrix c);

/* v := a * vb */
void _bncomp_mul_mpfmatrix_mpfvec(MPFVector v, MPFMatrix a, MPFVector vb);

/* v := a^T * vb */
void _bncomp_mul_mpfmatrixt_mpfvec(MPFVector v, MPFMatrix a, MPFVector vb);

/* initialize */
void _bncomp_init_g_mpf(int vec_dim, int mat_row_dim, int mat_col_dim);

/* finalize */
void _bncomp_free_g_mpf(void);

/* A^T * diag(d0, ..., dn) */
void _bncomp_mul_mpfmatrixt_mpfdiagmat(MPFMatrix ret, MPFMatrix a, MPFVector d);

/* substitute vector blocks */
void _bncomp_subst_mpfvector_blocks(MPFVector ret[], MPFVector v[], long int num_blocks);

/* inner product of vector blocks */
void _bncomp_ip_mpfvector_blocks(mpf_t ret, MPFVector va[], MPFVector vb[], long int num_blocks);

/* norm2 of vector blocks */
void _bncomp_norm2_mpfvector_blocks(mpf_t ret, MPFVector v[], long int num_blocks);

/* ret := va + alpha * vb */
void _bncomp_add_cmul_mpfvector_blocks(MPFVector ret[], MPFVector va[], mpf_t alpha, MPFVector vb[], long int num_blocks);

/* ret := (A \otimes I) vb[] */
void _bncomp_kmul_mpfmatrixI_mpfvector_blocks(MPFVector ret[], MPFMatrix mat, MPFVector vb[], long int num_blocks);

/* ret := -vb */
void _bncomp_neg_mpfvector_blocks(MPFVector vb[], long int num_blocks);

/* norm_inf of vector blocks */
void _bncomp_normi_mpfvector_blocks(mpf_t ret, MPFVector v[], long int num_blocks);

/* ret := va - vb */
void _bncomp_sub_mpfvector_blocks(MPFVector ret[], MPFVector va[], MPFVector vb[], long int num_blocks);

/* ret := va + vb */
void _bncomp_add_mpfvector_blocks(MPFVector ret[], MPFVector va[], MPFVector vb[], long int num_blocks);

/* ret := sum^n_{i,j=1} aij^2 */
void _bncomp_sumsqr_mpfmatrix(mpf_t ret, MPFMatrix mat);

/* Frobenius Norm for tridiagonal block matrix */
void _bncomp_normf_mpfmatrix_tridiag_blocks(mpf_t ret, MPFMatrix aim1[], MPFMatrix aii[], MPFMatrix aip1[], long int num_blocks);

/* Frobenius Norm for tridiagonal block band matrix */
void _bncomp_normf_mpfbmatrix_tridiag_blocks(mpf_t ret, MPFBMatrix aim1[], MPFBMatrix aii[], MPFBMatrix aip1[], long int num_blocks);

/* [ ret[0]              ]    [ b[0]              ]   [ aii [0] aip1[0]                                                   ]   [ x[0]              ] */
/* [ ret[1]              ]    [ b[1]              ]   [ aim1[0] aii [1] aip1[1]                                           ]   [ x[1]              ] */
/* [ ................... ] := [ ................. ] - [ ................................................................. ] * [ ................. ] */
/* [ ret[num_blocks - 2] ]    [ b[num_blocks - 2] ]   [    aim1[num_blocks - 3] aii [num_blocks - 2] aip1[num_blocsk - 2] ]   [ x[num_blocks - 2] ] */
/* [ ret[num_blocks - 1] ]    [ b[num_blocks - 1] ]   [                         aim1[num_blocks - 2] aii [num_blocks - 1] ]   [ x[num_blocks - 1] ] */
void _bncomp_residual_mpfmat_mpfvec_tridiag_blocks(MPFVector ret[], MPFVector b[], MPFMatrix aim1[], MPFMatrix aii[], MPFMatrix aip1[], MPFVector x[], long int num_blocks);

/* residual for band tridiagonal matrix */
void _bncomp_residual_mpfbmat_mpfvec_tridiag_blocks(MPFVector ret[], MPFVector b[], MPFBMatrix aim1[], MPFBMatrix aii[], MPFBMatrix aip1[], MPFVector x[], long int num_blocks);

/* mul vec */
/* ret[] := a[] * x[] */
void _bncomp_mul_mpfmat_mpfvec_tridiag_blocks(MPFVector ret[], MPFMatrix aim1[], MPFMatrix aii[], MPFMatrix aip1[], MPFVector x[], long int num_blocks);

/* mul vec : band */
/* ret[] := a[] * x[] */
void _bncomp_mul_mpfbmat_mpfvec_tridiag_blocks(MPFVector ret[], MPFBMatrix aim1[], MPFBMatrix aii[], MPFBMatrix aip1[], MPFVector x[], long int num_blocks);

/* ret[] := a[] * vb[] with working area */
void _bncomp_mul_mpftridiag_mpfvec_blocks(MPFVector ret[], MPFMatrix aim1[], MPFMatrix aii[], MPFMatrix aip1[], MPFVector vb[], MPFVector tmpv[], long int num_blocks);

/* ret[] := a_band[] * vb[] with working area */
void _bncomp_mul_mpfbtridiag_mpfvec_blocks(MPFVector ret[], MPFBMatrix aim1[], MPFBMatrix aii[], MPFBMatrix aip1[], MPFVector vb[], MPFVector tmpv[], long int num_blocks);

/* substitution of tridiagonal block matrix */
void _bncomp_subst_mpfmatrix_tridiag_blocks(MPFMatrix aim1_f[], MPFMatrix aii_f[], MPFMatrix aip1_f[], MPFMatrix aim1[], MPFMatrix aii[], MPFMatrix aip1[], long int num_blocks);

/* substitution of tridiagonal block matrix */
void _bncomp_subst_mpfbmatrix_tridiag_blocks(MPFBMatrix aim1_f[], MPFBMatrix aii_f[], MPFBMatrix aip1_f[], MPFBMatrix aim1[], MPFBMatrix aii[], MPFBMatrix aip1[], long int num_blocks);

/* substitute vector blocks */
void _bncomp_subst_dvector_mpfvec_blocks(DVector ret[], MPFVector v[], long int num_blocks);

/* substitute vector blocks */
void _bncomp_subst_mpfvector_dvec_blocks(MPFVector ret[], DVector v[], long int num_blocks);

/* substitution of tridiagonal block matrix */
void _bncomp_subst_dmatrix_mpfmat_tridiag_blocks(DMatrix aim1_f[], DMatrix aii_f[], DMatrix aip1_f[], MPFMatrix aim1[], MPFMatrix aii[], MPFMatrix aip1[], long int num_blocks);

/* substitution of tridiagonal block band matrix */
void _bncomp_subst_dbmatrix_mpfbmat_tridiag_blocks(DBMatrix aim1_f[], DBMatrix aii_f[], DBMatrix aip1_f[], MPFBMatrix aim1[], MPFBMatrix aii[], MPFBMatrix aip1[], long int num_blocks);

/* norm1 of vector blocks */
void _bncomp_norm1_mpfvector_blocks(mpf_t ret, MPFVector v[], long int num_blocks);

/* ret := vec / ||vec|| */
void _bncomp_normalize_mpfvector(mpf_t norm, MPFVector ret, MPFVector vec, int kind_of_norm);

/* vec := vec / ||vec|| */
void _bncomp_normalize2_mpfvector(mpf_t norm, MPFVector vec, int kind_of_norm);

/* ret := val * vec */
void _bncomp_cmul_mpfvector_blocks(MPFVector ret[], mpf_t val, MPFVector vec[], long int num_blocks);

/* vec := val * vec */
void _bncomp_cmul2_mpfvector_blocks(MPFVector vec[], mpf_t val, long int num_blocks);

/* vector norm */
void _bncomp_norm_mpfvector_blocks(mpf_t norm, MPFVector vec[], long int num_blocks, int kind_of_norm);

/* normalization: ret[] := vec[] / ||vec[]|| */
void _bncomp_normalize_mpfvector_blocks(mpf_t norm, MPFVector ret[], MPFVector vec[], long int num_blocks, int kind_of_norm);

/* normalization: vec[] /= ||vec[]|| */
void _bncomp_normalize2_mpfvector_blocks(mpf_t norm, MPFVector vec[], long int num_blocks, int kind_of_norm);

/* BiCGSTAB for band matrix */
long int _bncomp_MPFBBiCGSTAB_triblock(MPFVector answer[], MPFBMatrix aim1[], MPFBMatrix aii[], MPFBMatrix aip1[], MPFVector b[], mpf_t reps, mpf_t aeps, long int maxtimes, long int num_blocks);

/* BiCGSTAB */
long int _bncomp_MPFBiCGSTAB_triblock(MPFVector answer[], MPFMatrix aim1[], MPFMatrix aii[], MPFMatrix aip1[], MPFVector b[], mpf_t reps, mpf_t aeps, long int maxtimes, long int num_blocks);

/* Left preconditioned BiCGSTAB for band matrix */
long int _bncomp_MPFBBiCGSTAB_triblock_irk(MPFVector answer[], MPFBMatrix aim1[], MPFBMatrix aii[], MPFBMatrix aip1[], MPFBMatrix hmat[], MPFVector b[], mpf_t reps, mpf_t aeps, long int maxtimes, long int num_blocks);

/* Left preconditioned BiCGSTAB */
long int _bncomp_MPFBiCGSTAB_triblock_irk(MPFVector answer[], MPFMatrix aim1[], MPFMatrix aii[], MPFMatrix aip1[], MPFMatrix hmat[], MPFVector b[], mpf_t reps, mpf_t aeps, long int maxtimes, long int num_blocks);

/* GMRES(m) for band matrix */
int _bncomp_MPFBGMRESm_triblock(MPFVector ret_x[], MPFBMatrix aim1[], MPFBMatrix aii[], MPFBMatrix aip1[], MPFVector vec_b[], int restart_times, int maxtimes, mpf_t rtol, mpf_t atol, long int num_blocks);

/* GMRES(m) */
int _bncomp_MPFGMRESm_triblock(MPFVector ret_x[], MPFMatrix aim1[], MPFMatrix aii[], MPFMatrix aip1[], MPFVector vec_b[], int restart_times, int maxtimes, mpf_t rtol, mpf_t atol, long int num_blocks);

#endif // USE_GMP

#ifdef __cplusplus
} // extern "C"
#endif // __cplusplus

#endif // __BNCOMP_H
