/********************************************************************************/
/* ddlinear.h: Double-double and Quadruple precision Linear Computation Library */
/* Copyright (C) 2015 Tomonori Kouya                                            */
/*                                                                              */
/* This program is free software: you can redistribute it and/or modify it      */
/* under the terms of the GNU Lesser General Public License as published by the */
/* Free Software Foundation, either version 3 of the License or any later       */
/* version.                                                                     */
/*                                                                              */
/* This program is distributed in the hope that it will be useful, but WITHOUT  */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License */
/* for more details.                                                            */
/*                                                                              */
/* You should have received a copy of the GNU Lesser General Public License     */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                              */
/********************************************************************************/
// define __BNC_DDLINEAR_H__
#ifndef __BNC_DDLINEAR_H__
  #define __BNC_DDLINEAR_H__

#include "rdd.h"

#include "bnc.h"

#ifdef __cplusplus
extern "C" {
#endif

// DD vector
typedef struct
{
	long int dim;
	//double *element[DDSIZE]; // element[2]
#ifdef __cplusplus
	dd_real *element;
#else // __cplusplus
	double *element; // element[2]
#endif // __cplusplus
} ddvector;

typedef ddvector *DDVector;

// set_ddvector_i*
// get_ddvector_i*
#ifdef __cplusplus

	// C++ Macros
	#define SET_DDVECTOR_I(vec, index, value) {\
		vec->element[index] = value;\
	}
	#define SET_DDVECTOR_I_UI(vec, index, value) {\
		vec->element[index] = (double)value;\
	}
	#define SET_DDVECTOR_I_D(vec, index, value) {\
		vec->element[index] = (double)value;\
	}
	#define SET0_DDVECTOR_I(vec, index) {\
		vec->element[index] = (dd_real)0.0;\
	}

	#define GET_DDVECTOR_I(vec, index) (vec->element[index])

	#define set_ddvector_i(vec, index, value) SET_DDVECTOR_I(vec, index, value)
	#define set_ddvector_i_d(vec, index, value) SET_DDVECTOR_I_D(vec, index, value)
	#define set_ddvector_i_ui(vec, index, value) SET_DDVECTOR_I_UI(vec, index, value)
	#define set0_ddvector_i(vec, index) SET_DDVECTOR_I(vec, index)
	#define get_ddvector_i(vec, index) GET_DDVECTOR_I(vec, index)

#else // __cpplusplus

	// C Macros
	#define SET_DDVECTOR_I(vec, index, value) {\
		*(vec->element + (index) * DDSIZE) = value[0];\
		*(vec->element + (index) * DDSIZE + 1) = value[1];\
	}
	#define SET_DDVECTOR_I_UI(vec, index, value) {\
		*(vec->element + (index) * DDSIZE) = (double)value;\
		*(vec->element + (index) * DDSIZE + 1) = (double)0.0;\
	}
	#define SET_DDVECTOR_I_D(vec, index, value) {\
		*(vec->element + (index) * DDSIZE) = value;\
		*(vec->element + (index) * DDSIZE + 1) = (double)0.0;\
	}
	#define SET0_DDVECTOR_I(vec, index) {\
		*(vec->element + (index) * DDSIZE) = (double)0.0;\
		*(vec->element + (index) * DDSIZE + 1) = (double)0.0;\
	}

	#define GET_DDVECTOR_I(vec, index) (vec->element + (index) * DDSIZE)

	#ifndef USE_DDLINEAR_FUNCTIONS
		#define set_ddvector_i(vec, index, value) SET_DDVECTOR_I(vec, index, value)
		#define set_ddvector_i_d(vec, index, value) SET_DDVECTOR_I_D(vec, index, value)
		#define set_ddvector_i_ui(vec, index, value) SET_DDVECTOR_I_UI(vec, index, value)
		#define set0_ddvector_i(vec, index) SET_DDVECTOR_I(vec, index)
		#define get_ddvector_i(vec, index) GET_DDVECTOR_I(vec, index)
	#else
	// set element
	//void set_ddvector_i(DDVector vec, long int index, dd_real value)
	void set_ddvector_i(DDVector vec, long int index, double value[DDSIZE]);
	#endif // USE_DDLINEAR_FUNCTIONS

#endif // __cplusplus

#define gddvi get_ddvector_i
#define sddvi set_ddvector_i
#define gqdvi get_qdvector_i
#define sqdvi set_qdvector_i
#define gddmij get_ddmatrix_ij
#define sddmij set_ddmatrix_ij
#define gqdmij get_qdmatrix_ij
#define sqdmij set_qdmatrix_ij

// set a zero vector
void set0_ddvector(DDVector vec);

// initialize ddvector
DDVector init_ddvector(long int dim);

// free ddvector
void free_ddvector(DDVector vec);

// print ddvector
void print_ddvector(DDVector vec);

// set_ddvector_i_str
void set_ddvector_i_str(DDVector vec, long int index, const char *str);

/* c = a + b */
void add_ddvector(DDVector c, DDVector a, DDVector b);

/* c += a */
void add2_ddvector(DDVector c, DDVector a);

/* c = a - b */
void sub_ddvector(DDVector c, DDVector a, DDVector b);

/* c -= a */
void sub2_ddvector(DDVector c, DDVector a);

/* c = val * a */
#ifdef __cplusplus
void cmul_ddvector(DDVector c, dd_real val, DDVector a);
#else // __cplusplus
void cmul_ddvector(DDVector c, double val[DDSIZE], DDVector a);
#endif // __cplusplus

/* c *= val */
#ifdef __cplusplus
void cmul2_ddvector(DDVector c, dd_real val);
#else // __cplusplus
void cmul2_ddvector(DDVector c, double val[DDSIZE]);
#endif // __cplusplus

/* c = a + val * b */
#ifdef __cplusplus
void add_cmul_ddvector(DDVector c, DDVector a, dd_real val, DDVector b);
#else // __cplusplus
void add_cmul_ddvector(DDVector c, DDVector a, double val[DDSIZE], DDVector b);
#endif // __cplusplus

/* (a, b) */
#ifdef __cplusplus
void ip_ddvector(dd_real *ret, DDVector a, DDVector b);
#else // __cplusplus
void ip_ddvector(double ret[DDSIZE], DDVector a, DDVector b);
#endif // __cplusplus

/* c := a */
void subst_ddvector(DDVector c, DDVector a);

/* c := -a */
void neg_ddvector(DDVector c, DDVector a);

/* ||a||_1 */
#ifdef __cplusplus
void norm1_ddvector(dd_real *ret, DDVector a);
#else // __cplusplus
void norm1_ddvector(double ret[DDSIZE], DDVector a);
#endif // __cplusplus

/* ||a||_infty */
#ifdef __cplusplus
void normi_ddvector(dd_real *ret, DDVector a);
#else // __cplusplus
void normi_ddvector(double ret[DDSIZE], DDVector a);
#endif // __cplusplus

// Euclid norm
#ifdef __cplusplus
void norm2_ddvector(dd_real *ret, DDVector vec);
#else // __cplusplus
void norm2_ddvector(double ret[DDSIZE], DDVector vec);
#endif // __cplusplus

// QD vector
typedef struct
{
	long int dim;
	//double *element[QDSIZE]; // element[4]
#ifdef __cplusplus
	qd_real *element;
#else // __cplusplus
	double *element; // element[4]
#endif // __cplusplus
} qdvector;

typedef qdvector *QDVector;

// set_qdvector_i*
// get_qdvector_i*
#ifdef __cplusplus

	// C++ Macros 
	// Fix! 2015-04-25 by T.Kouya
	#define SET_QDVECTOR_I(vec, index, value) {\
		vec->element[index] = value;\
	}

	#define SET_QDVECTOR_I_UI(vec, index, value) {\
		vec->element[index] = (double)value;\
	}

	#define SET_QDVECTOR_I_D(vec, index, value) {\
		vec->element[index] = (double)value;\
	}

	#define SET0_QDVECTOR_I(vec, index) {\
		vec->element[index] = (qd_real)0.0;\
	}

	#define GET_QDVECTOR_I(vec, index) (vec->element[index])

	#define set_qdvector_i(vec, index, value) SET_QDVECTOR_I(vec, index, value)
	#define set_qdvector_i_d(vec, index, value) SET_QDVECTOR_I_D(vec, index, value)
	#define set_qdvector_i_ui(vec, index, value) SET_QDVECTOR_I_UI(vec, index, value)
	#define set0_qdvector_i(vec, index) SET0_QDVECTOR_I(vec, index)
	#define get_qdvector_i(vec, index) GET_QDVECTOR_I(vec, index)

#else // __cplusplus

	// C Macros
	#define SET_QDVECTOR_I(vec, index, value) { \
		*(vec->element + (index) * QDSIZE) = value[0];\
		*(vec->element + (index) * QDSIZE + 1) = value[1];\
		*(vec->element + (index) * QDSIZE + 2) = value[2];\
		*(vec->element + (index) * QDSIZE + 3) = value[3];\
	}

	#define SET_QDVECTOR_I_UI(vec, index, value) {\
		*(vec->element + (index) * QDSIZE) = (double)value;\
		*(vec->element + (index) * QDSIZE + 1) = (double)0.0;\
		*(vec->element + (index) * QDSIZE + 2) = (double)0.0;\
		*(vec->element + (index) * QDSIZE + 3) = (double)0.0;\
	}

	#define SET_QDVECTOR_I_D(vec, index, value) {\
		*(vec->element + (index) * QDSIZE) = value;\
		*(vec->element + (index) * QDSIZE + 1) = (double)0.0;\
		*(vec->element + (index) * QDSIZE + 2) = (double)0.0;\
		*(vec->element + (index) * QDSIZE + 3) = (double)0.0;\
	}

	#define SET0_QDVECTOR_I(vec, index) {\
		*(vec->element + (index) * QDSIZE) = (double)0.0;\
		*(vec->element + (index) * QDSIZE + 1) = (double)0.0;\
		*(vec->element + (index) * QDSIZE + 2) = (double)0.0;\
		*(vec->element + (index) * QDSIZE + 3) = (double)0.0;\
	}

	#define GET_QDVECTOR_I(vec, index) (vec->element + (index) * QDSIZE)

	#ifndef USE_QDLINEAR_FUNCTIONS
		#define set_qdvector_i(vec, index, value) SET_QDVECTOR_I(vec, index, value)
		#define set_qdvector_i_d(vec, index, value) SET_QDVECTOR_I_D(vec, index, value)
		#define set_qdvector_i_ui(vec, index, value) SET_QDVECTOR_I_UI(vec, index, value)
		#define set0_qdvector_i(vec, index) SET0_QDVECTOR_I(vec, index)
		#define get_qdvector_i(vec, index) GET_QDVECTOR_I(vec, index)
	#else
	// set element
	//void set_qdvector_i(DDVector vec, long int index, dd_real value)
	void set_qdvector_i(QDVector vec, long int index, double value[QDSIZE]);
	#endif // USE_QDLINEAR_FUNCTIONS

#endif // __cplusplus

// set a zero vector
//void set0_qdvector(QDVector vec)
void set0_qdvector(QDVector vec);

// initialize qdvector
QDVector init_qdvector(long int dim);

// free qdvector
void free_qdvector(QDVector vec);

// print ddvector
void print_qdvector(QDVector vec);

// set_qdvector_i_str
void set_qdvector_i_str(QDVector vec, long int index, const char *str);

/* c = a + b */
void add_qdvector(QDVector c, QDVector a, QDVector b);

/* c += a */
void add2_qdvector(QDVector c, QDVector a);

/* c = a - b */
void sub_qdvector(QDVector c, QDVector a, QDVector b);

/* c -= a */
void sub2_qdvector(QDVector c, QDVector a);

/* c = val * a */
#ifdef __cplusplus
void cmul_qdvector(QDVector c, qd_real val, QDVector a);
#else // __cplusplus
void cmul_qdvector(QDVector c, double val[QDSIZE], QDVector a);
#endif // __cplusplus

/* c *= val */
#ifdef __cplusplus
void cmul2_qdvector(QDVector c, qd_real val);
#else // __cplusplus
void cmul2_qdvector(QDVector c, double val[QDSIZE]);
#endif // __cplusplus

/* c = a + val * b */
#ifdef __cplusplus
void add_cmul_qdvector(QDVector c, QDVector a, qd_real val, QDVector b);
#else // __cplusplus
void add_cmul_qdvector(QDVector c, QDVector a, double val[QDSIZE], QDVector b);
#endif // __cplusplus

/* (a, b) */
#ifdef __cplusplus
void ip_qdvector(qd_real *ret, QDVector a, QDVector b);
#else // __cplusplus
void ip_qdvector(double ret[QDSIZE], QDVector a, QDVector b);
#endif // __cplusplus

/* c := a */
void subst_qdvector(QDVector c, QDVector a);

/* c := -a */
void neg_qdvector(QDVector c, QDVector a);

/* ||a||_1 */
#ifdef __cplusplus
void norm1_qdvector(qd_real *ret, QDVector a);
#else // __cplusplus
void norm1_qdvector(double ret[QDSIZE], QDVector a);
#endif // __cplusplus

/* ||a||_infty */
#ifdef __cplusplus
void normi_qdvector(qd_real *ret, QDVector a);
#else // __cplusplus
void normi_qdvector(double ret[QDSIZE], QDVector a);
#endif // __cplusplus

// Euclid norm
#ifdef __cplusplus
void norm2_qdvector(qd_real *ret, QDVector a);
#else // __cplusplus
void norm2_qdvector(double ret[QDSIZE], QDVector vec);
#endif // __cplusplus

// DD matrix
typedef struct
{
	long int row_dim, col_dim;
	//double *element[DDSIZE]; // element[2]
#ifdef __cplusplus
	dd_real *element;
#else // __cplusplus
	double *element; // element[2]
#endif // __cplusplus
} ddmatrix;

typedef ddmatrix *DDMatrix;

// set_ddmatrix_ij*
// get_ddmatrix_ij*
#ifdef __cplusplus 

	// C++ Macros
	#define SET_DDMATRIX_IJ(mat, row_index, col_index, value) {\
		mat->element[(row_index) * (mat->col_dim) + (col_index)] = value;\
	}
	#define SET_DDMATRIX_IJ_D(mat, row_index, col_index, value) {\
		mat->element[(row_index) * (mat->col_dim) + (col_index)] = (double)value;\
	}
	#define SET_DDMATRIX_IJ_UI(mat, row_index, col_index, value) {\
		mat->element[(row_index) * (mat->col_dim) + (col_index)] = (double)value;\
	}
	#define SET0_DDMATRIX_IJ(mat, row_index, col_index) {\
		mat->element[(row_index) * (mat->col_dim) + (col_index)] = (dd_real)0.0;\
	}

	#define GET_DDMATRIX_IJ(mat, row_index, col_index) (mat->element[(row_index) * (mat->col_dim) + (col_index)])

	#define set_ddmatrix_ij(mat, row_index, col_index, value) SET_DDMATRIX_IJ(mat, row_index, col_index, value)
	#define set_ddmatrix_ij_d(mat, row_index, col_index, value) SET_DDMATRIX_IJ_D(mat, row_index, col_index, value)
	#define set_ddmatrix_ij_ui(mat, row_index, col_index, value) SET_DDMATRIX_IJ_UI(mat, row_index, col_index, value)
	#define set0_ddmatrix_ij(mat, row_index, col_index) SET0_DDMATRIX_IJ(mat, row_index, col_index)
	#define get_ddmatrix_ij(mat, row_index, col_index) GET_DDMATRIX_IJ(mat, row_index, col_index)

#else // __cplusplus

	// C Macros
	#define SET_DDMATRIX_IJ(mat, row_index, col_index, value) {\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * DDSIZE) = value[0];\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * DDSIZE + 1) = value[1];\
	}
	#define SET_DDMATRIX_IJ_D(mat, row_index, col_index, value) {\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * DDSIZE) = value;\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * DDSIZE + 1) = (double)0.0;\
	}
	#define SET_DDMATRIX_IJ_UI(mat, row_index, col_index, value) {\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * DDSIZE) = (double)value;\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * DDSIZE + 1) = (double)0.0;\
	}
	#define SET0_DDMATRIX_IJ(mat, row_index, col_index) {\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * DDSIZE) = (double)0.0;\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * DDSIZE + 1) = (double)0.0;\
	}
		
	#define GET_DDMATRIX_IJ(mat, row_index, col_index) (mat->element + ((row_index) * mat->col_dim + (col_index)) * DDSIZE)

	#ifndef USE_DDLINEAR_FUNCTIONS
		#define set_ddmatrix_ij(mat, row_index, col_index, value) SET_DDMATRIX_IJ(mat, row_index, col_index, value)
		#define set_ddmatrix_ij_d(mat, row_index, col_index, value) SET_DDMATRIX_IJ_D(mat, row_index, col_index, value)
		#define set_ddmatrix_ij_ui(mat, row_index, col_index, value) SET_DDMATRIX_IJ_UI(mat, row_index, col_index, value)
		#define set0_ddmatrix_ij(mat, row_index, col_index) SET0_DDMATRIX_IJ(mat, row_index, col_index)
		#define get_ddmatrix_ij(mat, row_index, col_index) GET_DDMATRIX_IJ(mat, row_index, col_index)
	#else
	// set element
	//void set_ddvector_i(DDVector vec, long int index, dd_real value)
	void set_ddmatrix_ij(DDMatrix mat, long int row_index, long int col_index, double value[DDSIZE]);
	#endif // USE_DDLINEAR_FUNCTIONS

#endif // __cplusplus

// set a zero matrix
//void set0_ddmatrix(DDMatrix mat)
void set0_ddmatrix(DDMatrix mat);

// initialize ddvector
DDMatrix init_ddmatrix(long int row_dim, long int col_dim);

// free ddvector
void free_ddmatrix(DDMatrix mat);

// print ddvector
void print_ddmatrix(DDMatrix mat);

// matrix multiplication
// ret := A * B
void mul_ddmatrix(DDMatrix ret, DDMatrix a, DDMatrix b);

// Frobenius norm
#ifdef __cplusplus
void normf_ddmatrix(dd_real *ret, DDMatrix mat);
#else // __cplusplus
void normf_ddmatrix(double ret[DDSIZE], DDMatrix mat);
#endif // __cplusplus

// print normf
void print_normf_ddmatrix(const char *str, DDMatrix mat);

/* Infinity Norm of Matrix */
#ifdef __cplusplus
void normi_ddmatrix(dd_real *ret, DDMatrix mat);
#else // __cplusplus
void normi_ddmatrix(double ret[DDSIZE], DDMatrix mat);
#endif // __cplusplus

/* 1 Norm of Matrix */
#ifdef __cplusplus
void norm1_ddmatrix(dd_real *ret, DDMatrix mat);
#else // __cplusplus
void norm1_ddmatrix(double ret[DDSIZE], DDMatrix mat);
#endif // __cplusplus

/* c := a + b */
void add_ddmatrix(DDMatrix c, DDMatrix a, DDMatrix b);

/* c := a - b */
void sub_ddmatrix(DDMatrix c, DDMatrix a, DDMatrix b);

/* c := sc * a */
#ifdef __cplusplus
void cmul_ddmatrix_cpp(DDMatrix c, dd_real sc, DDMatrix a);
#else // __cplusplus
void cmul_ddmatrix(DDMatrix c, double sc[DDSIZE], DDMatrix a);
#endif // __cplusplus

/* c = a^T */
void transpose_ddmatrix(DDMatrix c, DDMatrix a);

/* c := a */
void subst_ddmatrix(DDMatrix c, DDMatrix a);

/* c := I */
void setI_ddmatrix(DDMatrix c);

/* v := a * vb */
void mul_ddmatrix_ddvec(DDVector v, DDMatrix a, DDVector vb);

/* v := a^T * vb */
void mul_ddmatrixt_ddvec(DDVector v, DDMatrix a, DDVector vb);

/* a = a^(-1) */
/* square matrix only */
void inv_ddmatrix(DDMatrix a);

// QD matrix
typedef struct
{
	long int row_dim, col_dim;
	//double *element[QDSIZE]; // element[4]
#ifdef __cplusplus
	qd_real *element;
#else // __cplusplus
	double *element; // element[4]
#endif // __cplusplus
} qdmatrix;

typedef qdmatrix *QDMatrix;

#ifdef __cplusplus
	// C++ Macros
	#define SET_QDMATRIX_IJ(mat, row_index, col_index, value) {\
		mat->element[(row_index) * (mat->col_dim) + (col_index)] = value;\
	}

	#define SET_QDMATRIX_IJ_D(mat, row_index, col_index, value) {\
		mat->element[(row_index) * (mat->col_dim) + (col_index)] = (double)value;\
	}

	#define SET_QDMATRIX_IJ_UI(mat, row_index, col_index, value) {\
		mat->element[(row_index) * (mat->col_dim) + (col_index)] = (double)value;\
	}

	#define SET0_QDMATRIX_IJ(mat, row_index, col_index) {\
		mat->element[(row_index) * (mat->col_dim) + (col_index)] = (qd_real)0.0;\
	}

	#define GET_QDMATRIX_IJ(mat, row_index, col_index) (mat->element[(row_index) * (mat->col_dim) + (col_index)])

	#ifndef USE_QDLINEAR_FUNCTIONS
		#define set_qdmatrix_ij(mat, row_index, col_index, value) SET_QDMATRIX_IJ(mat, row_index, col_index, value)
		#define set_qdmatrix_ij_d(mat, row_index, col_index, value) SET_QDMATRIX_IJ_D(mat, row_index, col_index, value)
		#define set_qdmatrix_ij_ui(mat, row_index, col_index, value) SET_QDMATRIX_IJ_UI(mat, row_index, col_index, value)
		#define set0_qdmatrix_ij(mat, row_index, col_index) SET0_QDMATRIX_IJ(mat, row_index, col_index)
		#define get_qdmatrix_ij(mat, row_index, col_index) GET_QDMATRIX_IJ(mat, row_index, col_index)
	#else
	// set element
	//void set_ddvector_i(DDVector vec, long int index, dd_real value)
	void set_qdmatrix_ij(QDMatrix mat, long int row_index, long int col_index, double value[QDSIZE]);
	#endif //  USE_QDLINEAR_FUNCTIONS


#else // __cplusplus

	// C Macros
	#define SET_QDMATRIX_IJ(mat, row_index, col_index, value) {\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE) = value[0];\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE + 1) = value[1];\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE + 2) = value[2];\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE + 3) = value[3];\
	}

	#define SET_QDMATRIX_IJ_D(mat, row_index, col_index, value) {\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE) = value;\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE + 1) = (double)0.0;\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE + 2) = (double)0.0;\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE + 3) = (double)0.0;\
	}

	#define SET_QDMATRIX_IJ_UI(mat, row_index, col_index, value) {\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE) = (double)value;\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE + 1) = (double)0.0;\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE + 2) = (double)0.0;\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE + 3) = (double)0.0;\
	}

	#define SET0_QDMATRIX_IJ(mat, row_index, col_index) {\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE) = (double)0.0;\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE + 1) = (double)0.0;\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE + 2) = (double)0.0;\
		*(mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE + 3) = (double)0.0;\
	}

	#define GET_QDMATRIX_IJ(mat, row_index, col_index) (mat->element + ((row_index) * mat->col_dim + (col_index)) * QDSIZE)

	#ifndef USE_QDLINEAR_FUNCTIONS
		#define set_qdmatrix_ij(mat, row_index, col_index, value) SET_QDMATRIX_IJ(mat, row_index, col_index, value)
		#define set_qdmatrix_ij_d(mat, row_index, col_index, value) SET_QDMATRIX_IJ_D(mat, row_index, col_index, value)
		#define set_qdmatrix_ij_ui(mat, row_index, col_index, value) SET_QDMATRIX_IJ_UI(mat, row_index, col_index, value)
		#define set0_qdmatrix_ij(mat, row_index, col_index) SET0_QDMATRIX_IJ(mat, row_index, col_index)
		#define get_qdmatrix_ij(mat, row_index, col_index) GET_QDMATRIX_IJ(mat, row_index, col_index)
	#else
	// set element
	//void set_ddvector_i(DDVector vec, long int index, dd_real value)
	void set_qdmatrix_ij(QDMatrix mat, long int row_index, long int col_index, double value[QDSIZE]);

	#endif //  USE_QDLINEAR_FUNCTIONS

#endif // __cplusplus

// set a zero matrix
void set0_qdmatrix(QDMatrix mat);

// initialize qdmatrix
QDMatrix init_qdmatrix(long int row_dim, long int col_dim);

// free qdmatrix
void free_qdmatrix(QDMatrix mat);

// print qdmatrix
void print_qdmatrix(QDMatrix mat);

// matrix multiplication
// ret := A * B
void mul_qdmatrix(QDMatrix ret, QDMatrix a, QDMatrix b);

// Frobenius norm
#ifdef __cplusplus
void normf_qdmatrix(qd_real *ret, QDMatrix mat);
#else // __cplusplus
void normf_qdmatrix(double ret[QDSIZE], QDMatrix mat);
#endif // __cplusplus

/* Infinity Norm of Matrix */
#ifdef __cplusplus
void normi_qdmatrix(qd_real *ret, QDMatrix mat);
#else // __cplusplus
void normi_qdmatrix(double ret[QDSIZE], QDMatrix mat);
#endif // __cplusplus

/* 1 Norm of Matrix */
#ifdef __cplusplus
void norm1_qdmatrix(qd_real *ret, QDMatrix mat);
#else // __cplusplus
void norm1_qdmatrix(double ret[QDSIZE], QDMatrix mat);
#endif // __cplusplus

/* c := a + b */
void add_qdmatrix(QDMatrix c, QDMatrix a, QDMatrix b);

/* c := a - b */
void sub_qdmatrix(QDMatrix c, QDMatrix a, QDMatrix b);

/* c := sc * a */
#ifdef __cplusplus
void cmul_qdmatrix(QDMatrix c, qd_real sc, QDMatrix a);
#else // __cplusplus
void cmul_qdmatrix(QDMatrix c, double sc[QDSIZE], QDMatrix a);
#endif // __cplusplus

/* c = a^T */
void transpose_qdmatrix(QDMatrix c, QDMatrix a);

/* c := a */
void subst_qdmatrix(QDMatrix c, QDMatrix a);

/* c := I */
void setI_qdmatrix(QDMatrix c);

/* v := a * vb */
void mul_qdmatrix_qdvec(QDVector v, QDMatrix a, QDVector vb);

/* v := a^T * vb */
void mul_qdmatrixt_qdvec(QDVector v, QDMatrix a, QDVector vb);

/* a = a^(-1) */
/* square matrix only */
void inv_qdmatrix(QDMatrix a);

#ifdef USE_GMP

// mpf to dd
#ifdef __cplusplus
void mpf_get_dd(dd_real *ret, mpf_t val);
//void mpf_get_dd(dd_real &ret, mpf_t val);
#else // __cplusplus
void mpf_get_dd(double ret[DDSIZE], mpf_t val);
#endif // __cplusplus

// dd to mpf
#ifdef __cplusplus
void mpf_set_dd(mpf_t ret, dd_real val);
#else // __cplusplus
void mpf_set_dd(mpf_t ret, double val[DDSIZE]);
#endif // __cplusplus

// mpf to qd
#ifdef __cplusplus
void mpf_get_qd(qd_real *ret, mpf_t val);
#else // __cplusplus
void mpf_get_qd(double ret[QDSIZE], mpf_t val);
#endif // __cplusplus

// qd to mpf
#ifdef __cplusplus
void mpf_set_qd(mpf_t ret, qd_real val);
#else // __cplusplus
void mpf_set_qd(mpf_t ret, double val[QDSIZE]);
#endif // __cplusplus

/* c := (mpf)a */
void subst_mpfvector_ddvec(MPFVector c, DDVector a);

/* c := (dd)a */
void subst_ddvector_mpfvec(DDVector c, MPFVector a);

/* c := (mpf)a */
void subst_mpfvector_qdvec(MPFVector c, QDVector a);

/* c := (qd)a */
void subst_qdvector_mpfvec(QDVector c, MPFVector a);

/* c := (mpf)a */
void subst_mpfmatrix_ddmat(MPFMatrix c, DDMatrix a);

/* c := (dd)a */
void subst_ddmatrix_mpfmat(DDMatrix c, MPFMatrix a);

/* c := (mpf)a */
void subst_mpfmatrix_qdmat(MPFMatrix c, QDMatrix a);

/* c := (qd)a */
void subst_qdmatrix_mpfmat(QDMatrix c, MPFMatrix a);

/* Normwise relative error of vector */
#ifdef __cplusplus
void relerr_ddvector_mpfvec(dd_real *relerr, DDVector approx_vec, MPFVector true_vec, int norm_type);
#else // __cplusplus
void relerr_ddvector_mpfvec(double relerr[DDSIZE], DDVector approx_vec, MPFVector true_vec, int norm_type);
#endif // __cplusplus

/* Elementwise relative errors of vector */
#ifdef __cplusplus
void relerr_element_ddvector_mpf(dd_real *max_relerr, dd_real *min_relerr, dd_real *norm_relerr, DDVector approx_vec, MPFVector true_vec, int norm_type);
#else // __cplusplus
void relerr_element_ddvector_mpf(double max_relerr[DDSIZE], double min_relerr[DDSIZE], double norm_relerr[DDSIZE], DDVector approx_vec, MPFVector true_vec, int norm_type);
#endif // __cplusplus

/* Normwise relative error of vector */
#ifdef __cplusplus
void relerr_qdvector_mpfvec(qd_real relerr, QDVector approx_vec, MPFVector true_vec, int norm_type);
#else // __cplusplus
void relerr_qdvector_mpfvec(double relerr[QDSIZE], QDVector approx_vec, MPFVector true_vec, int norm_type);
#endif // __cplusplus

/* Elementwise relative errors of vector */
#ifdef __cplusplus
void relerr_element_qdvector_mpf(qd_real &max_relerr, qd_real &min_relerr, qd_real &norm_relerr, QDVector approx_vec, MPFVector true_vec, int norm_type);
#else // __cplusplus
void relerr_element_qdvector_mpf(double max_relerr[QDSIZE], double min_relerr[QDSIZE], double norm_relerr[QDSIZE], QDVector approx_vec, MPFVector true_vec, int norm_type);
#endif // __cplusplus

#endif // USE_GMP

/* c := (dd)a */
void subst_ddvector_dvec(DDVector c, DVector a);

/* c := (d)a */
void subst_dvector_ddvec(DVector c, DDVector a);

/* c := (dd)a */
void subst_ddmatrix_dmat(DDMatrix c, DMatrix a);

/* c := (d)a */
void subst_dmatrix_ddmat(DMatrix c, DDMatrix a);

/* c := (qd)a */
void subst_qdvector_dvec(QDVector c, DVector a);

/* c := (d)a */
void subst_dvector_qdvec(DVector c, QDVector a);

/* c := (qd)a */
void subst_qdmatrix_dmat(QDMatrix c, DMatrix a);

/* c := (d)a */
void subst_dmatrix_qdmat(DMatrix c, QDMatrix a);

/* Normwise relative error of vector */
#ifdef __cplusplus
void relerr_ddvector(dd_real *relerr, DDVector approx_vec, DDVector true_vec, int norm_type);
#else // __cplusplus
void relerr_ddvector(double relerr[DDSIZE], DDVector approx_vec, DDVector true_vec, int norm_type);
#endif // __cplusplus

/* Elementwise relative errors of vector */
#ifdef __cplusplus
void relerr_element_ddvector(dd_real *max_relerr, dd_real *min_relerr, dd_real *norm_relerr, DDVector approx_vec, DDVector true_vec, int norm_type);
#else // __cplusplus
void relerr_element_ddvector(double max_relerr[DDSIZE], double min_relerr[DDSIZE], double norm_relerr[DDSIZE], DDVector approx_vec, DDVector true_vec, int norm_type);
#endif // __cplusplus

/* Normwise relative error of vector */
#ifdef __cplusplus
void relerr_qdvector(qd_real *relerr, QDVector approx_vec, QDVector true_vec, int norm_type);
#else // __cplusplus
void relerr_qdvector(double relerr[QDSIZE], QDVector approx_vec, QDVector true_vec, int norm_type);
#endif // __cplusplus

/* Elementwise relative errors of vector */
#ifdef __cplusplus
void relerr_element_qdvector(qd_real *max_relerr, qd_real *min_relerr, qd_real *norm_relerr, QDVector approx_vec, QDVector true_vec, int norm_type);
#else // __cplusplus
void relerr_element_qdvector(double max_relerr[QDSIZE], double min_relerr[QDSIZE], double norm_relerr[QDSIZE], QDVector approx_vec, QDVector true_vec, int norm_type);
#endif // __cplusplus

/**************************************/
/* ddlu.c                             */
/**************************************/
int DDLUdecomp(DDMatrix a);
int SolveDDLS(DDVector answer, DDMatrix lu, DDVector b);
int DDLUdecompP(DDMatrix a, long int ch[]);
int SolveDDLSP(DDVector answer, DDMatrix lu, DDVector b, long int ch[]);
int DDLUdecompC(DDMatrix a, long int row_ch[], long int col_ch[]);
int SolveDDLSC(DDVector answer, DDMatrix lu, DDVector b, long int row_ch[], long int col_ch[]);

int QDLUdecomp(QDMatrix a);
int SolveQDLS(QDVector answer, QDMatrix lu, QDVector b);
int QDLUdecompP(QDMatrix a, long int ch[]);
int SolveQDLSP(QDVector answer, QDMatrix lu, QDVector b, long int ch[]);
int QDLUdecompC(QDMatrix a, long int row_ch[], long int col_ch[]);
int SolveQDLSC(QDVector answer, QDMatrix lu, QDVector b, long int row_ch[], long int col_ch[]);

/**************************************/
/* ddbmatrix.c                        */
/**************************************/

// DD

typedef struct{
#ifdef __cplusplus
	dd_real *element;
	dd_real zero; // = 0
#else // __cplusplus
	double *element; // element[2]
	double zero[DDSIZE]; // = {0.0, 0.0}; // = 0
#endif // __cplusplus
	long int dim, upper_dim, lower_dim;
} ddbmatrix;

typedef ddbmatrix *DDBMatrix;

/* initialize band matrix */
DDBMatrix init_ddbmatrix(long int dim, long int upper_dim, long int lower_dim);

/* free mpfbmatrix */
void free_ddbmatrix(DDBMatrix mat);

/* return mat[i][j] */
double *get_ddbmatrix_ij(DDBMatrix mat, long int row_index, long int col_index);

/* set mat[i][j] = val */
void set_ddbmatrix_ij(DDBMatrix mat, long int row_index, long int col_index, double val[DDSIZE]);

/* set mat[i][j] = val */
void set_ddbmatrix_ij_d(DDBMatrix mat, long int row_index, long int col_index, double val);

/* set mat[i][j] = val */
void set_ddbmatrix_ij_ui(DDBMatrix mat, long int row_index, long int col_index, unsigned long val);

/* Multiply DDBMatrix * MPFVector */
int mul_ddbmatrix_ddvec(DDVector ret, DDBMatrix mat, DDVector vec);

/* Multiply DDBMatrix^T * MPFVector */
int mul_ddbmatrixt_ddvec(DDVector ret, DDBMatrix mat, DDVector vec);

int DDBLUdecomp(DDBMatrix a);

int SolveDDBLS(DDVector answer, DDBMatrix lu, DDVector b);

/* print band matrix */
void print_ddbmatrix(DDBMatrix mat);

/* set zero matrix */
void set0_ddbmatrix(DDBMatrix mat);

/* set identity matrix */
void setI_ddbmatrix(DDBMatrix mat);

/* Frobenius norm */
void normf_ddbmatrix(double ret[DDSIZE], DDBMatrix mat);

/* ret := mat_a + mat_b */
void add_ddbmatrix(DDBMatrix ret, DDBMatrix a, DDBMatrix b);

/* ret := mat_a - mat_b */
void sub_ddbmatrix(DDBMatrix ret, DDBMatrix a, DDBMatrix b);

/* ret := mat_a - mat_b */
void sub_ddmatrix_ddbmat_ddmat(DDMatrix ret, DDBMatrix a, DDMatrix b);

/* ret := val * a */
void cmul_ddbmatrix(DDBMatrix ret, double val[DDSIZE], DDBMatrix a);

/* ret := a * b */
void mul_ddmatrix_ddbmat_ddbmat(DDMatrix ret, DDBMatrix a, DDBMatrix b);

/* ret := a * b */
void mul_ddmatrix_ddmat_ddbmat(DDMatrix ret, DDMatrix a, DDBMatrix b);

/* ret := a * b */
void mul_ddmatrix_ddbmat_ddmat(DDMatrix ret, DDBMatrix a, DDMatrix b);

/* ret := a */
void subst_ddbmatrix(DDBMatrix ret, DDBMatrix a);

/* ret := (DBMatrix)a */
void subst_dbmatrix_ddbmat(DBMatrix ret, DDBMatrix a);

/* ret := (DDBMatrix)a */
void subst_ddbmatrix_dbmat(DDBMatrix ret, DBMatrix a);

/* ret := (DDMatrix)a */
void subst_ddmatrix_ddbmat(DDMatrix ret, DDBMatrix a);


/**************************************/
/* qdbmatrix.c                        */
/**************************************/
// QD

typedef struct{
#ifdef __cplusplus
	qd_real *element;
	qd_real zero; // = 0
#else // __cplusplus
	double *element; // element[2]
	double zero[QDSIZE]; // = {0.0, 0.0}; // = 0
#endif // __cplusplus
	long int dim, upper_dim, lower_dim;
} qdbmatrix;

typedef qdbmatrix *QDBMatrix;

/* initialize band matrix */
QDBMatrix init_qdbmatrix(long int dim, long int upper_dim, long int lower_dim);

/* free mpfbmatrix */
void free_qdbmatrix(QDBMatrix mat);

/* return mat[i][j] */
double *get_qdbmatrix_ij(QDBMatrix mat, long int row_index, long int col_index);

/* set mat[i][j] = val */
void set_qdbmatrix_ij(QDBMatrix mat, long int row_index, long int col_index, double val[QDSIZE]);

/* set mat[i][j] = val */
void set_qdbmatrix_ij_d(QDBMatrix mat, long int row_index, long int col_index, double val);

/* set mat[i][j] = val */
void set_qdbmatrix_ij_ui(QDBMatrix mat, long int row_index, long int col_index, unsigned long val);

/* Multiply QDBMatrix * MPFVector */
int mul_qdbmatrix_qdvec(QDVector ret, QDBMatrix mat, QDVector vec);

/* Multiply QDBMatrix^T * MPFVector */
int mul_qdbmatrixt_qdvec(QDVector ret, QDBMatrix mat, QDVector vec);

int QDBLUdecomp(QDBMatrix a);

int SolveQDBLS(QDVector answer, QDBMatrix lu, QDVector b);

/* print band matrix */
void print_qdbmatrix(QDBMatrix mat);

/* set zero matrix */
void set0_qdbmatrix(QDBMatrix mat);

/* set identity matrix */
void setI_qdbmatrix(QDBMatrix mat);

/* Frobenius norm */
void normf_qdbmatrix(double ret[QDSIZE], QDBMatrix mat);

/* ret := mat_a + mat_b */
void add_qdbmatrix(QDBMatrix ret, QDBMatrix a, QDBMatrix b);

/* ret := mat_a - mat_b */
void sub_qdbmatrix(QDBMatrix ret, QDBMatrix a, QDBMatrix b);

/* ret := mat_a - mat_b */
void sub_qdmatrix_qdbmat_qdmat(QDMatrix ret, QDBMatrix a, QDMatrix b);

/* ret := val * a */
void cmul_qdbmatrix(QDBMatrix ret, double val[QDSIZE], QDBMatrix a);

/* ret := a * b */
void mul_qdmatrix_qdbmat_qdbmat(QDMatrix ret, QDBMatrix a, QDBMatrix b);

/* ret := a * b */
void mul_qdmatrix_qdmat_qdbmat(QDMatrix ret, QDMatrix a, QDBMatrix b);

/* ret := a * b */
void mul_qdmatrix_qdbmat_qdmat(QDMatrix ret, QDBMatrix a, QDMatrix b);

/* ret := a */
void subst_qdbmatrix(QDBMatrix ret, QDBMatrix a);

/* ret := (DBMatrix)a */
void subst_dbmatrix_qdbmat(DBMatrix ret, QDBMatrix a);

/* ret := (QDBMatrix)a */
void subst_qdbmatrix_dbmat(QDBMatrix ret, DBMatrix a);

/* ret := (QDMatrix)a */
void subst_qdmatrix_qdbmat(QDMatrix ret, QDBMatrix a);


/**************************************/
/* fread_write_dd.c                   */
/**************************************/
void fread_ddmatrix(FILE *fp, DDMatrix mat);

void fread_ddmatrix(FILE *fp, DDMatrix mat);

void fread_ddmatrix_fname(const char *fname, DDMatrix mat);

void fwrite_ddmatrix(FILE *fp, DDMatrix mat);

void fwrite_ddmatrix_fname(const char *fname, DDMatrix mat);

//void fread_ddpolycoef(FILE *fp, DDPoly p, long int maxdeg);

//void fread_ddpolycoef_fname(const char *fname, DDPoly p, long int maxdeg);

void fread_ddvector(FILE *fp, DDVector vec);

void fread_ddvector_fname(const char *fname, DDVector vec);

void fwrite_ddvector(FILE *fp, DDVector vec);

void fwrite_ddvector_fname(const char *fname, DDVector vec);

/**************************************/
/* fread_write_qd.c                   */
/**************************************/
void fread_qdmatrix(FILE *fp, QDMatrix mat);

void fread_qdmatrix(FILE *fp, QDMatrix mat);

void fread_qdmatrix_fname(const char *fname, QDMatrix mat);

void fwrite_qdmatrix(FILE *fp, QDMatrix mat);

void fwrite_qdmatrix_fname(const char *fname, QDMatrix mat);

//void fread_qdpolycoef(FILE *fp, QDPoly p, long int maxdeg);

//void fread_qdpolycoef_fname(const char *fname, QDPoly p, long int maxdeg);

void fread_qdvector(FILE *fp, QDVector vec);

void fread_qdvector_fname(const char *fname, QDVector vec);

void fwrite_qdvector(FILE *fp, QDVector vec);

void fwrite_qdvector_fname(const char *fname, QDVector vec);


/**************************************/
/* gtestmat_dd.c                      */
/**************************************/
/* 1. Hilbert Matrix */
void hilbert_ddmatrix(DDMatrix a, long int dim);

/* 2. Lotkin Matrix */
void lotkin_ddmatrix(DDMatrix a, long int dim);

/* 3. Frank Matrix */
void frank_ddmatrix(DDMatrix a, long int dim);

/* 4. Tridiagonal Matrix */
void tridiag_ddmatrix(DDMatrix a, DDVector low_subdiag, DDVector diag, DDVector up_subdiag, long int dim);

/* 5. Integer Symmetrix Random Matrix */
void int_sym_rand_ddmatrix(DDMatrix mat, long int max, long int seed, long int dim);

/* 6. Integer Unsymmetrix Random Matrix */
void int_unsym_rand_ddmatrix(DDMatrix mat, long int max, long int seed, long int dim);

/* 7. Real Diagonal Matrix */
void diag_ddmatrix(DDMatrix mat, DDVector diag, long int dim);

/* 8. Toeplitz Matrix */
void toeplitz_ddmatrix(DDMatrix mat, double gamma_param[DDSIZE], long int dim);

/* 1. Hilbert Matrix */
void hilbert_qdmatrix(QDMatrix a, long int dim);

/* 2. Lotkin Matrix */
void lotkin_qdmatrix(QDMatrix a, long int dim);

/* 3. Frank Matrix */
void frank_qdmatrix(QDMatrix a, long int dim);

/* 4. Tridiagonal Matrix */
void tridiag_qdmatrix(QDMatrix a, QDVector low_subdiag, QDVector diag, QDVector up_subdiag, long int dim);

/* 5. Integer Symmetrix Random Matrix */
void int_sym_rand_qdmatrix(QDMatrix mat, long int max, long int seed, long int dim);

/* 6. Integer Unsymmetrix Random Matrix */
void int_unsym_rand_qdmatrix(QDMatrix mat, long int max, long int seed, long int dim);

/* 7. Real Diagonal Matrix */
void diag_qdmatrix(QDMatrix mat, QDVector diag, long int dim);

/* 8. Toeplitz Matrix */
void toeplitz_qdmatrix(QDMatrix mat, double gamma_param[QDSIZE], long int dim);

#ifdef __cplusplus
} // extern "C"
#endif

#endif // __BNC_DDLINEAR_H__
