/********************************************************************************/
/* matmulbench_dd.cc:                                                            */
/* Copyright (C) 2015 Tomonori Kouya                                            */
/*                                                                              */
/* This program is free software: you can redistribute it and/or modify it      */
/* under the terms of the GNU Lesser General Public License as published by the */
/* Free Software Foundation, either version 3 of the License or any later       */
/* version.                                                                     */
/*                                                                              */
/* This program is distributed in the hope that it will be useful, but WITHOUT  */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License */
/* for more details.                                                            */
/*                                                                              */
/* You should have received a copy of the GNU Lesser General Public License     */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                              */
/********************************************************************************/
#ifdef __cplusplus
	#include <cstdio>
	#include <cstdlib>
	#include <cmath>
	#include <iostream>
#else // __cplusplus
	#include <stdio.h>
	#include <stdlib.h>
	#include <math.h>
#endif // __cplusplus

#include "bnc.h"
#include "matmul_strassen.h"
#include "ddlinear.h"

#ifdef USE_OMP
	#include "bncomp.h"
#endif

// 256k bytes
//#define CACHESIZE (256 * 1024)

// 512k bytes
//#define CACHESIZE (512 * 1024)

// 1M bytes
//#define CACHESIZE (1 * 1024 * 1024)

// 2M bytes
//#define CACHESIZE (2 * 1024 * 1024)

void usage(const char *commandname)
{
#if defined(USE_STRASSEN) || defined(USE_STRASSEN_PADDING) || defined(USE_BLOCK)
	#ifdef USE_OMP
		printf("USAGE-> %s [min_dim] [num_threads] [Maximum seconds]\n", commandname);
	#else // USE_OMP
		printf("USAGE-> %s [min_dim] [Maximum seconds]\n", commandname);
	#endif // USE_OMP
#else
	#ifdef USE_OMP
		printf("USAGE-> %s [num_threads] [Maximum seconds]\n", commandname);
	#else // USE_OMP
		printf("USAGE-> %s [Maximum seconds]\n", commandname);
	#endif // USE_OMP
#endif
}

#ifdef USE_DDLINEAR


// relative errors
void relerr3_ddmatrix(dd_real *max_relerr, dd_real *min_relerr, dd_real *norm_relerr, DDMatrix mat, DDMatrix mat_true, int kind_of_norm)
{
	long int i, j, row_dim, col_dim;
	dd_real tmp_relerr, mat_true_norm;
	DDMatrix tmp_diff_mat;

	row_dim = mat->row_dim;
	col_dim = mat->col_dim;

	tmp_diff_mat = init_ddmatrix(row_dim, col_dim);

	// diff_mat := mat - mat_true
	sub_ddmatrix(tmp_diff_mat, mat, mat_true);

	// norm_relerr
	switch(kind_of_norm)
	{
		case 0: // infinit norm
			normi_ddmatrix(norm_relerr, tmp_diff_mat);
			normi_ddmatrix(&mat_true_norm, mat_true);
			break;
		case 1: // 1-norm
			norm1_ddmatrix(norm_relerr, tmp_diff_mat);
			norm1_ddmatrix(&mat_true_norm, mat_true);
			break;
		case 2: // Frobenius norm
		default:
			normf_ddmatrix(norm_relerr, tmp_diff_mat);
			normf_ddmatrix(&mat_true_norm, mat_true);
			break;
	}
	if(mat_true_norm != 0.0)
		*norm_relerr = *norm_relerr / mat_true_norm;

	// relative errors at each elements
	rdd_set_ui(*max_relerr, 0UL);
	rdd_set_ui(*min_relerr, 0UL);

	for(i = 0; i < row_dim; i++)
	{
		for(j = 0; j < col_dim; j++)
		{
			rdd_abs(tmp_relerr, get_ddmatrix_ij(tmp_diff_mat, i, j));
			if(get_ddmatrix_ij(mat_true, i, j) != 0.0)
			{
				rdd_div(tmp_relerr, tmp_relerr, get_ddmatrix_ij(mat_true, i, j));
				rdd_abs(tmp_relerr, tmp_relerr);
			}

			if(rdd_cmp(*max_relerr, tmp_relerr) < 0)
				rdd_set(*max_relerr, tmp_relerr);
			if(rdd_cmp(*min_relerr, tmp_relerr) > 0)
				rdd_set(*min_relerr, tmp_relerr);
		}
	}

	free_ddmatrix(tmp_diff_mat);

	return;
}

//int matmulloop_mpf(long int start_dim, long int end_dim, long int dim_step, long int maxsec, long int opt_dim_flag, unsigned long prec)
int matmulloop_dd(long int start_dim, long int end_dim, long int dim_step, long int maxsec, long int opt_dim_flag, long int min_dim)
{
	long int dim;
	long int i, j, end_flag;
	long int num_addsub, num_mul;
	double iteration, etime, stime, mflops;
	DDMatrix da, db, dc;
	//DDMatrix dc_long;
	DDMatrix dc_long;
	dd_real tmp, relerr[3];

	for(dim = start_dim; dim <= end_dim; dim += dim_step)
	{
		da = init_ddmatrix(dim, dim);
		db = init_ddmatrix(dim, dim);
		dc = init_ddmatrix(dim, dim);
		//dc_long = init_ddmatrix(dim, dim);
		dc_long = init_ddmatrix(dim, dim);

		for(i = 0; i < dim; i++)
		{
			for(j = 0; j < dim; j++)
			{
//				set_ddmatrix_ij(da, i, j, (i + j + 1) * sqrt(5.0));
				rdd_sqrt_ui(tmp, 5UL);
				rdd_mul_ui(tmp, tmp, (unsigned long)(i + j + 1));
				set_ddmatrix_ij(da, i, j, tmp);

//				set_ddmatrix_ij(db, i, j, (dim - i) * sqrt(3.0));
				rdd_sqrt_ui(tmp, 3UL);
				rdd_mul_ui(tmp, tmp, (unsigned long)(dim - i));
				set_ddmatrix_ij(db, i, j, tmp);

				rdd_sqrt_ui(tmp, 2UL);
				set_ddmatrix_ij(dc, i, j, tmp);
			}
		}
		//rdd_sqrt_ui(tmp, 5UL);
		//std::cout << "sqrt5 = " << tmp.to_string() << std::endl;
		//rdd_mul_ui(tmp, tmp, (unsigned long)(2 + 3 + 1));
		//std::cout << "5 / 6 = " << tmp.to_string() << std::endl;

		//normf_ddmatrix(&tmp, da); std::cout << "||A||_F = " << tmp << std::endl;
		//normf_ddmatrix(&tmp, da); std::cout << "||B||_F = " << tmp << std::endl;

		iteration = 1.0;
		do{
			stime = get_real_secv();
#ifdef USE_STRASSEN
			for(i = 0; i < iteration; i++)
			{
#ifdef USE_OMP
				_bncomp_reset_num_mul_ddmatrix_strassen();
//				mul_ddmatrix_strassen(dc, da, db, _BNC_DEFAULT_MIN_DIM_STRASSEN);
				_bncomp_mul_ddmatrix_strassen(dc, da, db, min_dim);
//				_bncomp_mul_ddmatrix_strassen_nonrec(dc, da, db, min_dim);
				_bncomp_get_num_mul_ddmatrix_strassen(&num_addsub, &num_mul);
#else // USE_OMP
				reset_num_mul_ddmatrix_strassen();
//				mul_ddmatrix_strassen(dc, da, db, _BNC_DEFAULT_MIN_DIM_STRASSEN);
				mul_ddmatrix_strassen(dc, da, db, min_dim);
				get_num_mul_ddmatrix_strassen(&num_addsub, &num_mul);
#endif // USE_OMP
			}

#elif USE_STRASSEN_PADDING
			for(i = 0; i < iteration; i++)
			{
#ifdef USE_OMP
				_bncomp_reset_num_mul_ddmatrix_strassen();
//				mul_ddmatrix_strassen(dc, da, db, _BNC_DEFAULT_MIN_DIM_STRASSEN);
				_bncomp_mul_ddmatrix_strassen(dc, da, db, min_dim);
				_bncomp_get_num_mul_ddmatrix_strassen(&num_addsub, &num_mul);
#else // USE_OMP
				reset_num_mul_ddmatrix_strassen();
//				mul_ddmatrix_strassen_odd_padding(dc, da, db, _BNC_DEFAULT_MIN_DIM_STRASSEN);
				mul_ddmatrix_strassen_odd_padding(dc, da, db, min_dim);
				get_num_mul_ddmatrix_strassen(&num_addsub, &num_mul);
#endif // USE_OMP
			}

#elif USE_BLOCK
			num_addsub = da->row_dim * dc->row_dim * dc->col_dim;
			num_mul = da->row_dim * dc->row_dim * dc->col_dim;
#ifdef USE_OMP
//			for(i = 0; i < iteration; i++) mul_ddmatrix_block(dc, da, db, _BNC_DEFAULT_MIN_DIM_STRASSEN);
			for(i = 0; i < iteration; i++) _bncomp_mul_ddmatrix_block(dc, da, db, min_dim);
#else // USE_OMP
//			for(i = 0; i < iteration; i++) mul_ddmatrix_block(dc, da, db, _BNC_DEFAULT_MIN_DIM_STRASSEN);
			for(i = 0; i < iteration; i++) mul_ddmatrix_block(dc, da, db, min_dim);
#endif // USE_OMP
#else
			num_addsub = da->row_dim * dc->row_dim * dc->col_dim;
			num_mul = da->row_dim * dc->row_dim * dc->col_dim;
#ifdef USE_OMP
			for(i = 0; i < iteration; i++) _bncomp_mul_ddmatrix_simple(dc, da, db);
#else // USE_OMP
			for(i = 0; i < iteration; i++) mul_ddmatrix_simple(dc, da, db);
#endif // USE_OMP
#endif
			//etime = get_secv() - stime;
			etime = get_real_secv() - stime;
			iteration *= 2.0;
		} while(etime < 2);

		// get relative errors
#ifdef USE_OMP
		_bncomp_mul_ddmatrix_simple(dc_long, da, db);
#else // USE_OMP
//		mul_ddmatrix(dc_long, da, db, _BNC_DEFAULT_MIN_DIM_STRASSEN);
		mul_ddmatrix_simple(dc_long, da, db);
#endif // USE_OMP
		relerr3_ddmatrix(&relerr[0], &relerr[1], &relerr[2], dc, dc_long, 0);

		etime /= iteration / 2;

//		mflops = (double)dim * (double)dim * (double)dim * 2 / etime / 1000.0 / 1000.0;
		mflops = (double)(num_addsub + num_mul) / etime / 1000.0 / 1000.0;
		//printf("[%5d bits] %5d x %5d: %9.5f sec (%8.5f micro-sec/mul) %f MFlops", prec, dim, dim, etime, (etime / dim / dim / dim) * 1000 * 1000, mflops);
		//printf("[%5d bits] %5d x %5d: %9.5f sec (%8.5f micro-sec/mul) %f MFlops", prec, dim, dim, etime, (etime / (double)num_mul) * 1000 * 1000, mflops);
		printf("[%5d bits] %5d x %5d: %9.5f sec [mul]%ld [addsub]%ld", 106, dim, dim, etime, num_mul, num_addsub);
		printf(" "); printf("%10.3e", to_double(relerr[0])); //rdd_out_str(relerr[0]);
		printf(" "); printf("%10.3e", to_double(relerr[1])); //rdd_out_str(relerr[1]);
		printf(" "); printf("%10.3e", to_double(relerr[2])); //rdd_out_str(relerr[2]);// printf("\n");

		if(opt_dim_flag == 1)
			printf(" <-- Optimal dimension %+d", dim - (end_dim + start_dim) / 2);
		printf("\n");

		free_ddmatrix(da);
		free_ddmatrix(db);
		free_ddmatrix(dc);
		free_ddmatrix(dc_long);

		// exceed max. comp. time
		if(etime > maxsec)
		{
			end_flag = 1;
			break;
		}
	}

	return end_flag;
}
#endif

int main(int argc, char *argv[])
{
	int num_threads = 1;
	long int dim, idim, idim_half, opt_dim, end_flag, min_dim = _BNC_DEFAULT_MIN_DIM_STRASSEN;
	long int cachesize = 128, maxsec = 600;

	if(argc < 1)
	{
		usage(argv[0]);
		return 0;
	}
	else if(argc >= 2)
	{
#if defined(USE_STRASSEN) || defined(USE_STRASSEN_PADDING) || defined(USE_BLOCK)
		min_dim = atol(argv[1]);
		if(min_dim < _BNC_DEFAULT_MIN_DIM_STRASSEN)
		{
			printf("Warning: Too small min_dim (%ld < %ld)\n", min_dim, _BNC_DEFAULT_MIN_DIM_STRASSEN);
			//return -1;
		}
#ifdef USE_OMP
		if(argc >= 3)
		{
			num_threads = atoi(argv[2]);
			if(num_threads < 1)
			{
				printf("num_threads is 1\n");
			}
		}
		if(argc >= 4)
		{
			maxsec = atol(argv[3]);
			if(maxsec < 0)
			{
				printf("Invalid Maximum seconds! (%ld seconds)\n", maxsec);
				return -1;
			}
		}
#else // USE_OMP
		maxsec = atol(argv[1]);
		if(maxsec < 0)
		{
			printf("Invalid Maximum seconds! (%ld seconds)\n", maxsec);
			return -1;
		}
#endif // USE_OMP
#else // defined(USE_STRASSEN) || defined(USE_STRASSEN_PADDING) || defined(USE_BLOCK)
#ifdef USE_OMP
		num_threads = atoi(argv[1]);
		if(num_threads < 1)
		{
			printf("num_threads is 1\n");
		}
		if(argc >= 3)
		{
			maxsec = atol(argv[2]);
			if(maxsec < 0)
			{
				printf("Invalid Maximum seconds! (%ld seconds)\n", maxsec);
				return -1;
			}
		}
#else // USE_OMP
		maxsec = atol(argv[1]);
		if(maxsec < 0)
		{
			printf("Invalid Maximum seconds! (%ld seconds)\n", maxsec);
			return -1;
		}
#endif // USE_OMP
#endif // defined(USE_STRASSEN) || defined(USE_STRASSEN_PADDING) || defined(USE_BLOCK)
	}

	printf("------------------ Benchmark Test using Matrix Multiplication -----------------\n");

	// Initialize QD library
	fpu_fix_start(NULL);
	printf("Maxsec: %ld\n", maxsec);

#ifdef USE_OMP
	set_bncomp_num_threads(num_threads);
	printf("OpenMP #threads = %d\n", omp_get_num_threads());
#endif

//	for(idim = 32; idim <= 16384; idim *= 2)
	//for(idim = 2; idim <= 2048; idim *= 2)
	for(idim = min_dim; idim < 32768; idim *= 2)
	{
		end_flag = 0;

		end_flag = matmulloop_dd(idim - 1, idim + 1, 1, maxsec, 0, min_dim);

//		end_flag = matmulloop_dd(idim - 1, idim + 1, 1, maxsec, 0);
		idim_half = 3 * idim / 2;

		if(idim_half >= min_dim)
			end_flag = matmulloop_dd(idim_half - 1, idim_half + 1, 1, maxsec, 0, min_dim);

//		if((idim <= opt_dim) && (opt_dim <= idim * 2))
//			end_flag = matmulloop_dd(opt_dim - 5, opt_dim + 5, 1, maxsec, 1, min_dim);

		if(end_flag == 1)
			break;

	}
	printf("------------------------------------ END --------------------------------------\n");
}

